// Copyright (C) 2006, 2007, 2008, 2009 by Florent Lamiraux, Thomas
// Moulard, Mathieu Poirier, CNRS.
//
// This file is part of the hppWalkPlanner.
//
// hppWalkPlanner is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or (at your option) any later version.
//
// hppWalkPlanner is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with hppWalkPlanner. If not, see
// <http://www.gnu.org/licenses/>.

#ifndef HPP_WALKPLANNER_PLANNER_HH
# define HPP_WALKPLANNER_PLANNER_HH
# include <vector>

# include <hpp/util/portability.hh>
# include <hpp/walkplanner/fwd.hh>
# include <hpp/walkfootplanner/planner.hh>

namespace hpp
{
  namespace walkPlanner
  {
    /**
       \brief Implement a two step walk planning strategy.

       The motion planning strategy is based on an iterative method that
       plan successive motions. In a first step, a path is computed for
       the bounding box of the robot and then this path is animated using
       JRL walk pattern generator. This iterative strategy uses two
       instanciations of the humanoid robot:

       \li a bounding box and
       \li a fully actuated humanoid robot.

       <b> How to use this class</b>

       To use this class,

       \li create an object Planner, by specifying the step size,
       \li insert a humanoid robot using ChppPlanner::addHppProblem in the
       vector of problems,
       \li put this robot in the starting configuration of a walk motion,
       \li call Planner::initializeProblem to build the bounding
       box of the humanoid robot,
       \li add obstacles using ChppPlanner::addObstacle

       \li choose initial and goal configurations of the box using
       Planner::initConfig and Planner::goalConfig and

       \li call Planner::solve to compute a collision-free walking
       motion for the humanoid robot.

       To get the resulting path, use ChppPlanner::getPath with
       problemId=1 and pathId = ChppPlanner::getNbPaths(1)

       Note that the default steering method for the bounding box is
       ChppWalkSteeringMethod, but it can be changed using ChppPlanner
       API.

    */
    class HPP_DLLAPI Planner : public walkFootPlanner::Planner
    {
    public:
      /**
	 \brief Constructor

	 \param inStepLength Abscissa of the landing foot in the frame
	 of the support foot in straight line motions.

	 \param inStepSpace Absolute value of the ordinate of the
	 landing foot in the frame of the support foot

	 \param inSecurityDistance Distance above which centers of
	 feet should remain in two consecutive steps.

	 \param inBoxLateralTolerance lateral growing of bounding box
	 to include swinging motion during walk.

	 \param inModel Path to VRML model of HRP2.

      */
      Planner (double inStepLength,
	       double inStepSpace,
	       double inSecurityDistance,
	       double inBoxLateralTolerance);

      /// \brief Destructor
      virtual ~Planner ();

      /// \name Problem definition
      /// @{

      /**
	 \brief Set initial configuration of Walk Planning problem
	 \param x position and orientation of the robot.
	 \param y position and orientation of the robot.
	 \param theta position and orientation of the robot.

	 Call ChppFootPlanner::initConfig() and set humanoid robot in corresponding
	 half-sitting configuration;
      */
      ktStatus initConfig (double x, double y, double theta);

      /**
	 \brief Set goal configuration of Walk Planning problem
	 \param x position and orientation of the robot.
	 \param y position and orientation of the robot.
	 \param theta position and orientation of the robot.

	 Call ChppFootPlanner::goalConfig() and set humanoid robot in corresponding
	 half-sitting configuration;
      */
      ktStatus goalConfig (double x, double y, double theta);

      /**
	 \brief Initialize a Walk Planning problem.

	 Compute the bounding box of the first robot in problem vector
	 (in given configuration) and create a robot with this
	 box. Then insert this new robot at beginning of problem
	 vector.

	 The box robot is placed in the first problem by a push_front().

      */
      virtual ktStatus initializeProblem ();

      /// @}


      /// \name Problem resolution
      /// @{

      /**
	 \brief Iteratively solve a walk motion planning problem for a
	 humanoid robot.

	 \return KD_OK or KD_ERROR
      */
      virtual ktStatus solve ();

      bool
	writeSeqplayFiles
	(const std::vector<CkwsConfigShPtr>& path,
	 const std::vector<double>& walkTime,
	 const std::vector<std::vector<double> >& zmp,
	 const std::vector<std::vector<double> >& waist) const;

      /**
	 \brief Get the all the element to compute the solution on
	 HRP2.

	 \param outHRP2Path the hpp path of HRP2

	 \param outHRP2TimeRef the Time Reference for all the element

	 \param outHRP2ZMPRef a vector of ZMPRef(3) ... vector resize
	 on TimeRef size

	 \param outHRP2WaistPosition a vector of WaistPosition(6)
	 ... vector resize on TimeRef size

	 \return KD_OK or KD_ERROR
      */
      ktStatus solution
      (std::vector<CkwsConfigShPtr> &outHRP2Path,
       std::vector<double> &outHRP2TimeRef,
       std::vector<std::vector<double> > &outHRP2ZMPRef,
       std::vector<std::vector<double> > &outHRP2WaistPosition );

      /**
	 \brief compute the 2st phase : compute all angle for hrp2

	 \return KD_OK or KD_ERROR
      */
      ktStatus computeHumanoidTrajectory ();
      /// @}

    private:
      /// \brief Reset Trajectory of humanoid robot
      void resetHumanoidTrajectory()
      {
	attHRP2Path.clear ();
	attHRP2TimeRef.clear ();
	attHRP2ZMPRef.clear ();
	attHRP2WaistPosition.clear ();
      }

      /// The compute Path of HRP2
      std::vector<CkwsConfigShPtr> attHRP2Path;

      /// The time Reference of HRP2 path
      std::vector<double> attHRP2TimeRef;

      /// The ZMP reference of HRP2 Path
      std::vector<std::vector<double> > attHRP2ZMPRef;

      /// the Waist Position of HRP2 Path
      std::vector<std::vector<double> > attHRP2WaistPosition;

      /// \brief Object that compute the stack of steps and the robot walk trajectory.
      PatternGenerator* attHppPatGen;

      /**
	 \brief Lateral tolerance of the bounding box

	 The bounding box of the robot is enlarged by this value.
      */
      double attBoxLateralTolerance;

    };

  } // end of namespace walkPlanner.
} // end of namespace hpp.

#endif //! HPP_WALKPLANNER_PLANNER_HH
