// Copyright (C) 2006, 2007, 2008, 2009 by Florent Lamiraux, Thomas
// Moulard, Mathieu Poirier, CNRS.
//
// This file is part of the hppWalkPlanner.
//
// hppWalkPlanner is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or (at your option) any later version.
//
// hppWalkPlanner is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with hppWalkPlanner.  If not, see
// <http://www.gnu.org/licenses/>.

#ifndef HPPWALKPLANNER_PATTERNGENERATOR_HH
# define HPPWALKPLANNER_PATTERNGENERATOR_HH
# include <iostream>
# include <fstream>

# include <KineoWorks2/kwsPath.h>
# include <hppModel/hppDevice.h>

# include <hpp/walkfootplanner/footstep-sequence.hh>

# include "walkgenjrl-nowarning.hh"

namespace hpp
{
  namespace walkPlanner
  {

    /// \brief API to use the walkGenJRL library for collision free
    /// walk planning
    class HPP_DLLAPI PatternGenerator
    {
    public:
      typedef hpp::walkFootPlanner::footStep_t footStep_t;
      typedef hpp::walkFootPlanner::FootStepSequence footStepSequence_t;

      /**
	 \brief constructor that sets the robot in half-sitting
	 position.

	 \param inStepLength Abscissa of the landing foot in the frame
	 of the support foot in straight line motions.

	 \param inStepSpace Absolute value of the ordinate of the
	 landing foot in the frame of the support foot

	 \param inModel Path to VRML model of HRP2.

      */
      PatternGenerator(double inStepLength,
		       double inStepSpace);

      /// \brief  destructor
      ~PatternGenerator ();

      /// \brief Read only access to step spacing
      double stepSpacing () const
      {
	return attStepSpacing;
      }

      /// \brief Reset the configuration of the robot in half-sitting
      /// position.
      void reset ();

      /**
	 \brief Initialize a PatternGenerator (interface) with several
	 variables

	 \li samplingperiod : 0.005
	 \li previewcontroltime : 1.6
	 \li comheight : 0.8078
	 \li omega  : 0.0
	 \li stepheight : 0.07
	 \li singlesupporttime : 0.78
	 \li doublesupporttime : 0.02
	 \li armparameters :  0.5
	 \li LimitsFeasibility : 0.0
	 \li ZMPShiftParameters :  0.015 0.015 0.015 0.015
	 \li TimeDistributionParameters : 2.0 3.7 1.7 3.0
	 \li UpperBodyMotionParameters :  -0.1 -1.0 0.0

      */
      void initWalkPatGen (CjrlHumanoidDynamicRobot* inRobot);

      /**
	 \brief compute all the Q configuration in 5 ms time for
	 OpenHRP (in FILE .pos, ...) and KINEOWORKS (in structure
	 pathHrp2)

	 \param hppDevice kwsDevice of HRP2

	 \param outPathHumanoid kwsPath of HRP2 to be play in the
	 kppInterface

	 \param outDiscretizedPath vector of HRP2 kwsConfigs equal to
	 HRP2 kwsPath for external use as OpenHrp

	 \param outHRP2TimeRef vector of Time reference for
	 outHRP2Path outHRP2ZMPRef and outHRP2WaistPosition

	 \param outHRP2ZMPRef vector or ZMP trajectory corresponding
	 to the outHRP2Path

	 \param outHRP2WaistPosition vector or Waist trajectory
	 (position and orientation) corresponding to the outHRP2Path

	 \param inRelativeFootPrintVector step sequence

	 The trajectory of the robot is moved by
	 (PatternGenerator::m_xInit PatternGenerator::m_yInit
	 PatternGenerator::m_thetaInit). See
	 PatternGenerator::computeAStepSequenceFromAPath for
	 explanation.
      */
      ktStatus
	computeAllQFromAStepSequence
	(ChppDeviceShPtr hppDevice,
	 CkwsPathShPtr& outPathHumanoid,
	 std::vector<CkwsConfigShPtr>& outDiscretizedPath,
	 std::vector<double>& outHRP2TimeRef,
	 std::vector<std::vector<double> >& outHRP2ZMPRef,
	 std::vector<std::vector<double> >& outHRP2WaistPosition,
	 const footStepSequence_t& inRelativeFootPrintVector);

      /**
	 \brief compute a stream for Any Curve : step seq :
	 ................. from a step Sequence for
	 PatternGeneratorInterface

	 \param inRelativeFootPrintVector step sequence.
      */
      void
	computeStreamFromStepSequence
	(const footStepSequence_t& inRelativeFootPrintVector);

    private:

      /// \brief Length of the steps
      ///
      /// Abscissa of the landing foot in the frame of the support
      /// foot in straight line motions.
      double attStepLength;

      /// \brief Space between feet
      ///
      /// Absolute value of the ordinate of the landing foot in the
      /// frame of the support foot
      double attStepSpacing;

      /// \brief String stream passed to pattern generator interface
      std::istringstream attStringStream;

      /// \brief Interface with walk patter generator
      PatternGeneratorJRL::PatternGeneratorInterface* attWalkPatGen;
    };

  } // end of namespace walkPlanner.
} // end of namespace hpp.

#endif //! HPPWALKPLANNER_PATTERNGENERATOR_HH
