// Copyright (C) 2006, 2007, 2008, 2009 by Florent Lamiraux, Thomas
// Moulard, Mathieu Poirier, CNRS.
//
// This file is part of the hppWalkPlanner.
//
// hppWalkPlanner is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or (at your option) any later version.
//
// hppWalkPlanner is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with hppWalkPlanner. If not, see
// <http://www.gnu.org/licenses/>.

#include "config.h"
#include "kdunix-warning.hh"

#include <iostream>
#include <fstream>

#include <KineoModel/kppSolidComponentRef.h>
#include <KineoWorks2/kwsPathOptimizer.h>
#include <KineoWorks2/kwsRoadmapBuilder.h>
#include <KineoWorks2/kwsConfig.h>

#include <hrp2_14/hrp2_14.h>

#include <hpp/util/debug.hh>
#include <hpp/util/timer.hh>
#include "hpp/walkplanner/planner.hh"

#include "hppModel/hppHumanoidRobot.h"

#include "pattern-generator.hh"

namespace hpp
{
  namespace walkPlanner
  {
    Planner::Planner
    (double inStepLength, double inStepSpace,
     double inFootSecurityDistance,
     double inBoxLateralTolerance)
      : walkFootPlanner::Planner (inStepLength,
				  inStepSpace,
				  inFootSecurityDistance),
	attHppPatGen (new PatternGenerator (inStepLength, inStepSpace)),
	attBoxLateralTolerance (inBoxLateralTolerance)
    {
      assert (attHppPatGen);

      attHRP2Path.clear ();
      attHRP2TimeRef.clear ();
      attHRP2ZMPRef.clear ();
      attHRP2WaistPosition.clear ();
    }

    Planner::~Planner ()
    {
      attHRP2Path.clear ();
      attHRP2TimeRef.clear ();
      attHRP2ZMPRef.clear ();
      attHRP2WaistPosition.clear ();

      delete attHppPatGen;
    }

    ktStatus
    Planner::initializeProblem ()
    {
      double xMin=0, yMin=0, zMin=0, xMax=0, yMax=0, zMax=0;
      unsigned nbRobots = getNbHppProblems ();

      if (nbRobots < 1)
	{
	  hppDout (error, "missing robot");
	  return KD_ERROR;
	}

      // Assume humanoid robot is at rank 0 in problem vector.
      ChppHumanoidRobotShPtr humanoidRobot = KIT_DYNAMIC_PTR_CAST
	(ChppHumanoidRobot, robotIthProblem (0));

      // Initialize interface to walk pattern generator
      attHppPatGen->initWalkPatGen(humanoidRobot.get());

      // Compute bouding box of humanoid robot.
      if (humanoidRobot->axisAlignedBoundingBox
	  (xMin, yMin, zMin, xMax, yMax, zMax) != KD_OK)
	{
	  hppDout (error, "missing robot geometry");
	  return KD_ERROR;
	}

      const ChppDeviceShPtr& robotBox =
	createRobotBox(xMin, yMin - attBoxLateralTolerance, -0.01,
		       xMax, yMax + attBoxLateralTolerance, zMax);

      if (!robotBox)
	{
	  hppDout (error, "failed to create bounding box");
	  return KD_ERROR;
	}

      //make the joint invisible by default
      robotBox->isVisible (false);

      //the body of the bouding box transparent
      robotBox->solidComponentRef (0)
	->referencedSolidComponent ()->isTransparent (true);

      // Set same default penetration for box and humanoid
      const double boxPenetration = penetration (0);

      // Register bounding box in problem vector.
      addHppProblemAtBeginning (robotBox, boxPenetration);

      // desactived HRP2 and box collision
      robotBox->ignoreDeviceForCollision (humanoidRobot);
      humanoidRobot->ignoreDeviceForCollision (robotBox);

      if (ChppWalkFootPlanner::initializeProblem () != KD_OK)
	return KD_ERROR;
      return KD_OK;
    }

    ktStatus
    Planner::initConfig(double inX, double inY, double inTheta)
    {
      if (ChppWalkFootPlanner::initConfig(inX, inY, inTheta) != KD_OK)
	return KD_ERROR;

      if (getNbHppProblems() < 2)
	{
	  hppDout (error, "missing some robots (only "
		   << getNbHppProblems () << " has/have been defined)");
	  return KD_ERROR;
	}

      // Build configuration for HRP2.
      const ChppDeviceShPtr HRP2Device =
	KIT_DYNAMIC_PTR_CAST(ChppDevice, robotIthProblem (1));

      if( !HRP2Device)
	hppDout (warning, "empty HRP-2") ;


      CkwsConfigShPtr initConfHRP2 = CkwsConfig::create (HRP2Device);
      initConfHRP2->dofValue (0, inX);
      initConfHRP2->dofValue (1, inY);
      initConfHRP2->dofValue (2, 0);
      initConfHRP2->dofValue (3, 0);
      initConfHRP2->dofValue (4, 0);
      initConfHRP2->dofValue (5, inTheta);

      if (!initConfHRP2)
	{
	  hppDout (error, "could not create HRP-2 initial configuration");
	  return KD_ERROR;
	}

      if (initConfIthProblem(1, initConfHRP2) == KD_ERROR)
	{
	  hppDout (error, "could not set HRP-2 initial configuration");
	  return KD_ERROR;
	}

      return KD_OK;
    }

    ktStatus
    Planner::goalConfig(double inX, double inY, double inTheta)
    {
      if (ChppWalkFootPlanner::goalConfig(inX, inY, inTheta) != KD_OK)
	return KD_ERROR;

      if (getNbHppProblems() < 2)
	{
	  hppDout (error, "missing some robots (only "
		   << getNbHppProblems () << " has/have been defined)");
	  return KD_ERROR;
	}

      // Build configuration for HRP2.
      const ChppDeviceShPtr HRP2Device =
	KIT_DYNAMIC_PTR_CAST(ChppDevice, robotIthProblem (1));

      if(!HRP2Device)
	hppDout (warning, "empty HRP-2");

      CkwsConfigShPtr goalConfHRP2 =
	CkwsConfig::create(HRP2Device);
      goalConfHRP2->dofValue (0, inX);
      goalConfHRP2->dofValue (1, inY);
      goalConfHRP2->dofValue (2, 0);
      goalConfHRP2->dofValue (3, 0);
      goalConfHRP2->dofValue (4, 0);
      goalConfHRP2->dofValue (5, inTheta);

      if (!goalConfHRP2)
	{
	  hppDout (error, "failed to create HRP-2 goal configuration");
	  return KD_ERROR;
	}

      if (goalConfIthProblem(1, goalConfHRP2) == KD_ERROR)
	{
	  hppDout (error, "failed to set HRP-2 goal configuration");
	  return KD_ERROR;
	}

      return KD_OK;
    }


    ktStatus
    Planner::solve()
    {
      if (ChppWalkFootPlanner::solve () == KD_ERROR)
	return KD_ERROR;
      hppDout (info, "__________ PHASE 3: Humanoid trajectory __________" );
      return computeHumanoidTrajectory ();
    }

    bool
    Planner::writeSeqplayFiles
    (const std::vector<CkwsConfigShPtr>& path,
     const std::vector<double>& walkTime,
     const std::vector<std::vector<double> >& zmp,
     const std::vector<std::vector<double> >& waist) const
    {
      assert (path.size () >= 1);
      std::ofstream posFile
	(debug::getFilename ("trajectory.pos", PACKAGE_TARNAME).c_str ());
      std::ofstream rpyFile
	(debug::getFilename ("trajectory.rpy", PACKAGE_TARNAME).c_str ());
      std::ofstream zmpFile
	(debug::getFilename ("trajectory.zmp", PACKAGE_TARNAME).c_str ());

      // Get HRP-2 device.
      const ChppDeviceShPtr hrp2 =
	KIT_DYNAMIC_PTR_CAST(ChppDevice, robotIthProblem (1));
      assert (hrp2);

      std::vector<double> kwsConfigDofVector;
      vectorN jrlDynamicsDofVector (hrp2->numberDof ());

      for (unsigned walkIndex = 0; walkIndex < walkTime.size (); ++walkIndex)
	{
	  posFile << walkTime[walkIndex] << " ";
	  rpyFile << walkTime[walkIndex] << " ";
	  zmpFile << walkTime[walkIndex] << " ";

	  if (path[walkIndex]->getDofValues (kwsConfigDofVector) != KD_OK)
	    return false;

	  // Convert Kineo vector of dof values into
	  // OpenHRP configuration vector.
	  if (!hrp2->kwsToJrlDynamicsDofValues
	      (kwsConfigDofVector, jrlDynamicsDofVector))
	    return false;

	  // The first 6 degrees of freedom represent waist configuration.
	  for(unsigned i = 6; i < jrlDynamicsDofVector.size (); ++i)
	    posFile << jrlDynamicsDofVector[i] << " ";
	  posFile << "\n";

	  rpyFile << waist[walkIndex][3] << " "
		  << waist[walkIndex][4] << " "
		  << waist[walkIndex][5] << "\n";

	  zmpFile << zmp[walkIndex][0] << " "
		  << zmp[walkIndex][1] << " "
		  << zmp[walkIndex][2] << "\n";
	}
      return true;
    }

    ktStatus
    Planner::computeHumanoidTrajectory()
    {
      ktStatus status = KD_OK;

      CkwsPathShPtr pathHrp2 = CkwsPath::create (robotIthProblem (1));
      attHppPatGen->reset ();
      resetHumanoidTrajectory ();
      hppDout (info, "clear all vector");

      hppStartBenchmark (HUMANOID_TRAJECTORY);
      if (attHppPatGen->computeAllQFromAStepSequence
	  (KIT_DYNAMIC_PTR_CAST(ChppDevice, robotIthProblem (1)),
	   pathHrp2, attHRP2Path,
	   attHRP2TimeRef, attHRP2ZMPRef, attHRP2WaistPosition,
	   footStepSequence ()))
	{
	  hppDout (warning, "detected collision during walking motion");
	  if (pathHrp2)
	    addPath(1, pathHrp2);
	  status = KD_ERROR;
	}
      hppStopBenchmark (HUMANOID_TRAJECTORY);
      hppDisplayBenchmark (HUMANOID_TRAJECTORY);

      if (status == KD_ERROR)
	return KD_ERROR;

      assert (pathHrp2 && "No returned path.");

      // put the REAL Config in init and goal configs of HRP2
      initConfIthProblem (1, pathHrp2->configAtStart ());
      goalConfIthProblem (1, pathHrp2->configAtEnd ());

      hppDout (info, "Number of direct paths in path: "
	       << pathHrp2->countDirectPaths ());
      addPath(1,pathHrp2);

      hppDebug (writeSeqplayFiles (attHRP2Path,
				   attHRP2TimeRef,
				   attHRP2ZMPRef,
				   attHRP2WaistPosition));

      return KD_OK;
    }


    ktStatus
    Planner::solution
    (std::vector<CkwsConfigShPtr>& outHRP2Path,
     std::vector<double>& outHRP2TimeRef,
     std::vector<std::vector<double> >& outHRP2ZMPRef,
     std::vector<std::vector<double> >& outHRP2WaistPosition)
    {

      if( attHRP2Path.empty ())
	{
	  hppDout (error, "missing HRP-2 path");
	  return KD_ERROR;
	}

      outHRP2Path = attHRP2Path;

      if (attHRP2TimeRef.empty ())
	{
	  hppDout (error, "missing time ref");
	  return KD_ERROR;
	}
      outHRP2TimeRef = attHRP2TimeRef;

      if (attHRP2ZMPRef.empty())
	{
	  hppDout (error, "missing HRP-2 ZMP ref");
	  return KD_ERROR;
	}
      outHRP2ZMPRef = attHRP2ZMPRef;

      if(attHRP2WaistPosition.empty())
	{
	  hppDout (error, "missing HRP-2 waist positions");
	  return KD_ERROR;
	}
      outHRP2WaistPosition = attHRP2WaistPosition ;
      return KD_OK ;
    }

  } // end of namespace walkPlanner.
} // end of namespace hpp.
